import os
import fnmatch
from pathlib import Path

from .main import remove_prefix


ABSTRA_FOLDER = ".abstra/"
ABSTRA_IGNORE = ".abstraignore"
CREDENTIALS_FILE = ".abstra/credentials"


def get_ignore_files(dir):
    IGNOREFILE = os.path.join(dir, ABSTRA_IGNORE)
    abstra_path = os.path.join(dir, ABSTRA_FOLDER)
    ignored = [IGNOREFILE, abstra_path]
    if os.path.exists(IGNOREFILE):
        with open(IGNOREFILE, "r") as f:
            ignored.extend([os.path.join(dir, f) for f in f.read().split("\n") if f])
    return ignored


def should_ignore(ignored_paths, _path):
    path = str(_path)
    for _ignored_path in ignored_paths:
        ignored_path = normalize_path(_ignored_path)
        if fnmatch.fnmatch(path, ignored_path):
            return True
        if fnmatch.fnmatch(path, ignored_path + "/*"):
            return True
    return False


def normalize_path(path):
    path = str(path)
    if path.endswith("/"):
        path = path[:-1]
    if path.startswith("./"):
        path = path[2:]
    return path


def files_from_directory(directory):
    ignored = [*get_ignore_files(directory), *get_ignore_files(os.getcwd())]
    paths = Path(directory).rglob("*")
    paths = [
        path for path in paths if path.is_file() and not should_ignore(ignored, path)
    ]
    return paths


def remove_filepath_prefix(filepath, prefix):
    return remove_prefix(normalize_path(filepath), normalize_path(prefix) + "/")



import webbrowser


from abstra_cli.resources.resources import Resource
import abstra_cli.messages as messages
import abstra_cli.utils as utils
import abstra_cli.apis as apis
import os
import json
from glob import glob


NAME_PARAMETERS = ["name", "title"]
PATH_PARAMETERS = ["path"]
CODE_PARAMETERS = ["code", "c"]
BACKGROUND_PARAMETERS = ["background"]
OTHER_PARAMETERS = [
    "main_color",
    "font_family",
    "logo_url",
    "log_messages",
    "brand_name",
]
LAYOUT_PARAMETERS = ["layout"]
NON_FLAG_PARAMETERS = (
    NAME_PARAMETERS
    + PATH_PARAMETERS
    + CODE_PARAMETERS
    + BACKGROUND_PARAMETERS
    + OTHER_PARAMETERS
    + LAYOUT_PARAMETERS
)
FLAG_PARAMETERS = ["auto_start", "allow_restart", "show_sidebar", "enabled"]
DASH_PARAMETERS = FLAG_PARAMETERS + NON_FLAG_PARAMETERS


def check_valid_parameters(parameters: dict) -> None:
    for param in parameters.keys():
        if param not in DASH_PARAMETERS:
            messages.invalid_parameter(param)
            exit()
    for param, value in parameters.items():
        if param in NON_FLAG_PARAMETERS and value in [True, False]:
            messages.invalid_non_flag_parameter_value(param)
            exit()


def evaluate_parameter_name(parameters: dict, use_default=True) -> dict:
    name = parameters.get("name") or parameters.get("n") or parameters.get("title")
    if not name and not use_default:
        return {}
    return {"name": name or "New Dash"}


def evaluate_parameter_path(parameters: dict) -> dict:
    path = parameters.get("path")
    if not path:
        return {}
    return {"path": path}


def evaluate_parameters_code(parameters: dict, use_default=True) -> dict:
    EMPTY_DASH = ""
    code = parameters.get("code") or parameters.get("c")

    if not code:
        raise Exception("Code is required")
    return {"code_file_path": code}


def evaluate_parameter_layout(parameters: dict) -> dict:
    layout = parameters.get("layout")
    if not layout:
        raise Exception("Layout is required")
    return {"layout": layout}


def evaluate_flag_parameters(parameters: dict) -> dict:
    evaluated_params = {}
    for param, value in parameters.items():
        if param in FLAG_PARAMETERS:
            if value == "true" or value == True:
                evaluated_params[param] = True
                continue
            if value == "false" or value == False:
                evaluated_params[param] = False
                continue

            messages.invalid_flag_parameter_value(param)
            exit()
    return evaluated_params


def evaluate_other_parameters(parameters: dict) -> dict:
    other_parameters = {}
    for param, value in parameters.items():
        if param in OTHER_PARAMETERS:
            other_parameters[param] = value

    return other_parameters


def evaluate_background_parameter_value(parameters: dict) -> dict:
    background = parameters.get("background", None)
    if not background:
        return {}

    if utils.check_color(background):
        return {"theme": background}

    if not utils.path_exists(background):
        messages.file_path_does_not_exists_message(background)
        exit()

    if utils.check_is_image_path(background):
        filename = utils.slugify_filename(background)
        try:
            with open(background, "rb") as f:
                file = f.read()
                url = apis.asset_upload(filename, file)
                return {"theme": url}
        except Exception as e:
            messages.error_upload_background_message(background)
            exit()

    messages.invalid_background_parameter_value()
    exit()


class Dashes(Resource):
    @staticmethod
    def list():
        dashes = apis.list_workspace_dashes()
        messages.print_dashes(dashes)

    @staticmethod
    def add(*args, **kwargs):
        upsert = kwargs.pop("upsert", False)
        path = kwargs.get("path")
        if upsert and not path:
            messages.upsert_without_identifier("path")
            exit()

        check_valid_parameters(kwargs)

        dash_data = {
            **evaluate_parameter_name(kwargs),
            **evaluate_parameter_path(kwargs),
            **evaluate_parameters_code(kwargs),
            **evaluate_flag_parameters(kwargs),
            **evaluate_other_parameters(kwargs),
            **evaluate_background_parameter_value(kwargs),
            **evaluate_parameter_layout(kwargs),
        }

        if dash_data:
            try:
                if upsert:
                    apis.upsert_workspace_dash(dash_data)
                    messages.upserted_message("Dash", path)
                else:
                    path = apis.add_workspace_dash(dash_data)["path"]
                    messages.created_message("Dash", path)
            except Exception as e:
                print(e)
                messages.create_failed("Dash")

    @staticmethod
    def update(*args, **kwargs):
        if not len(args):
            messages.required_argument("path")
            exit()
        path = args[0]

        if not len(kwargs):
            messages.missing_parameters_to_update("dash", path)
            exit()

        check_valid_parameters(kwargs)

        dash_data = {
            **evaluate_parameter_name(kwargs, use_default=False),
            **evaluate_parameter_path(kwargs),
            **evaluate_parameters_code(kwargs, use_default=False),
            **evaluate_flag_parameters(kwargs),
            **evaluate_other_parameters(kwargs),
            **evaluate_background_parameter_value(kwargs),
            **evaluate_parameter_layout(kwargs),
        }

        if dash_data:
            try:
                apis.update_workspace_dash(path, dash_data)
                messages.updated_message("Dash", path)
            except Exception as e:
                print(e)
                messages.update_failed("Dash", path)

    @staticmethod
    def remove(*args, **kwargs):
        if not len(args):
            messages.required_argument("path")
            exit()

        path = args[0]
        apis.delete_workspace_dash(path)
        messages.deleted_message("Dash", path)

    @staticmethod
    def play(*args, **kwargs):
        if not len(args):
            messages.required_argument("path")
            exit()

        path = args[0]
        subdomain_name = apis.get_subdomain()
        url = utils.get_prod_dash_url(subdomain_name, path)
        messages.dash_url(url)
        webbrowser.open(url)

    @staticmethod
    def list_dash_props(abstra_json_path):
        abstra_json_dir = os.path.dirname(abstra_json_path)
        with open(abstra_json_path, "r") as f:
            workspace_json_data = json.load(f)
        root_dir = workspace_json_data.get("workspace", {"root": "."})["root"]
        dash_files = glob(
            os.path.join(abstra_json_dir, root_dir, "**", "*.abstradash.json"),
            recursive=True,
        )
        dash_props = []
        for dash_file_path in dash_files:
            common_path = dash_file_path.replace(".abstradash.json", "")
            route = os.path.relpath(
                common_path, os.path.join(abstra_json_dir, root_dir)
            ).replace("\\", "/")
            script_path = common_path + ".py"
            dash_json_data = json.load(open(dash_file_path, "r"))
            prop = {
                "name": dash_json_data["name"],
                "layout": dash_json_data["layout"],
                "background": workspace_json_data["workspace"]["theme"],
                "path": route,
                "code": script_path,
            }
            dash_props.append(prop)
        return dash_props



import os
from pathlib import Path

from abstra_cli.resources.resources import Resource
import abstra_cli.messages as messages
import abstra_cli.utils as utils
import abstra_cli.apis as apis


class Files(Resource):
    @staticmethod
    def list():
        files = apis.list_workspace_files()
        messages.print_files(files)

    @staticmethod
    def add(*args, **kwargs):
        files: list[Path] = []
        for path in args:
            if os.path.isfile(path):
                files.append(Path(path))
            elif os.path.isdir(path):
                files.extend(utils.files_from_directory(path))

        bar = messages.show_progress("Uploading files", len(files))
        for path in files:
            filename = path.as_posix()
            ok = apis.upload_file(filename, path.open("rb"))
            if not ok:
                print(f"Error uploading file {filename}")
                return False
            else:
                bar.next()
        bar.finish()
        print(f"\nUploaded {len(files)} files successfully")

    @staticmethod
    def remove(*args, **kwargs):
        # TODO: list first then delete
        bar = messages.show_progress("Deleting files", len(args))
        for f in args:
            ok = apis.delete_file(f)
            if not ok:
                print(f"Error deleting file {f}")
                return False
            else:
                bar.next()
        bar.finish()
        print(f"\nDeleted {len(args)} files successfully")



import abstra_cli.apis.main as api_main
import abstra_cli.utils as utils


def list_workspace_jobs():
    query = """
        query GetJobs {
            jobs {
                title
                schedule
                identifier
                script {
                    enabled
                }
            }
        }
    """
    return api_main.hf_hasura_runner(query).get("jobs", [])


def add_workspace_job(data):
    _, workspace_id, _ = api_main.get_auth_info()
    job_data = {
        "title": data["name"],
        "script": {
            "data": {
                "workspace_id": workspace_id,
                "enabled": data.get("enabled", True),
                "code": data["code"],
                "name": data["name"],
            }
        },
    }

    data.pop("name", None)
    data.pop("code", None)
    data.pop("enabled", None)
    job_data.update(data)

    query = """
        mutation InsertJob($job_data: [jobs_insert_input!]!) {
            insert_jobs(
                objects: $job_data
            ) {
                returning {
                    identifier
                    title
                }
            }
        }
    """

    return (
        api_main.hf_hasura_runner(query, {"job_data": job_data})
        .get("insert_jobs", {})
        .get("returning", {})[0]
    )


def update_workspace_job(identifier, data):
    job_data = data.copy()
    script_data = {}

    name = job_data.pop("name", None)
    if name:
        job_data["title"] = name
        script_data["name"] = name

    code = job_data.pop("code", None)
    if code:
        script_data["code"] = code

    enabled = job_data.pop("enabled", None)
    if enabled is not None:
        script_data["enabled"] = enabled

    request_data = {
        "identifier": identifier,
        "job_data": job_data,
        "script_data": script_data,
    }
    update_query = """
        mutation UpdateJob($identifier: String!, $job_data: jobs_set_input, $script_data: scripts_set_input = {}) {
            update_jobs(where: {identifier: {_eq: $identifier}}, _set: $job_data) {
                returning {
                    id
                    title
                    identifier
                }
            }
            update_scripts(where: {jobs: {identifier: {_eq: $identifier}}}, _set: $script_data) {
                returning {
                    name
                }
            }
        }
    """
    return api_main.hf_hasura_runner(update_query, request_data)


def upsert_workspace_job(data):
    identifier = data["identifier"]

    query = """
        query FindJob($identifier: String!) {
            jobs(where: {identifier: {_eq: $identifier}}) {
                identifier
            }
        }
    """

    jobs = api_main.hf_hasura_runner(query, {"identifier": identifier}).get("jobs")
    if len(jobs):
        return update_workspace_job(identifier, data)
    else:
        return add_workspace_job(data)


def delete_workspace_job(identifier):
    query = """
        mutation DeleteJob($identifier: String!) {
            delete_jobs(where: {identifier: {_eq: $identifier}}) {
                returning {
                    id
                    identifier
                    title
                }
            }
        }
    """

    return api_main.hf_hasura_runner(query, {"identifier": identifier})



import os, re
from colour import Color


IMAGE_EXTENSIONS = [".png", ".jpg", ".jpeg", ".svg"]
SPACE = " "
EMPTY = ""
DASH = "-"
DOT = "."

to_lower = lambda strs: list(map(lambda s: s.lower(), strs))
to_kebab = lambda strs: f"{DASH}".join(to_lower(strs))
reverse = lambda strs: strs[::-1]


def remove_prefix(text, prefix):
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


def digits(n):
    return len(str(n))


def format_digits(n, d):
    return " " * (d - digits(n)) + str(n)


def parse_env_var(var):
    match = re.search("^(\w+)=(.+)$", var.strip())
    if match:
        return match.group(1), match.group(2)
    return None, None


def parse_package(pkg):
    # https://peps.python.org/pep-0440/#version-specifiers

    match = re.search(
        "^(([\w\.\-]+)\s*)(~=|===|>=|<=|==)(\s*([^\[\]\,\;\>\<\s]+))$", pkg
    )
    if match:
        name = match.group(2)
        version = match.group(5)
        return name, version

    if re.search("\>|\<|\,|\;|\[|\]", pkg):
        return None, None

    return pkg, None


def check_color(color: str) -> bool:
    try:
        color = color.replace(SPACE, EMPTY) if isinstance(color, str) else None
        Color(color)
        return True
    except ValueError:
        return False


def slugify_filename(image_path: str) -> str:
    filename = image_path.split("/")[-1]
    extension, *filenames = reverse(filename.split("."))
    return f"{DOT}".join(reverse([extension, to_kebab(filenames)]))


def check_is_image_path(image_path: str) -> bool:
    _, file_extension = os.path.splitext(image_path)
    return file_extension in IMAGE_EXTENSIONS


path_exists = lambda file_path: os.path.exists(file_path)

get_prod_form_url = (
    lambda subdomain_name, path: f"https://{subdomain_name}.abstra.run/{path}"
)

get_prod_dash_url = (
    lambda subdomain_name, path: f"https://{subdomain_name}.abstra.run/{path}"
)

get_prod_hook_url = (
    lambda subdomain_name, path: f"https://hooks.abstra.cloud/{subdomain_name}/{path}"
)



import fire

from abstra_cli.deploy import deploy
import abstra_cli.messages as messages
import abstra_cli.decorators as decorators
import abstra_cli.credentials as credentials
from abstra_cli.resources import Forms, Files, Packages, Vars, Hooks, Jobs, Dashes


class CLI(object):
    """
    A CLI to manage your Abstra Cloud Workspace environment.

    usage: abstra <command> <resource> [<argument> ...] [parameters]
    """

    @decorators.configuration_check
    def configure(self, api_token=None):
        credentials.save_credentials(api_token or messages.read_credentials())
        print("Done!")

    @decorators.credentials_check
    def list(self, resource, **kwargs):
        list_func = {
            "vars": Vars.list,
            "jobs": Jobs.list,
            "files": Files.list,
            "forms": Forms.list,
            "hooks": Hooks.list,
            "dashes": Dashes.list,
            "packages": Packages.list,
        }.get(resource, messages.not_implemented)

        list_func()

    @decorators.credentials_check
    def add(self, resource, *args, **kwargs):
        add_func = {
            "vars": Vars.add,
            "dash": Dashes.add,
            "dashes": Dashes.add,
            "form": Forms.add,
            "forms": Forms.add,
            "hook": Hooks.add,
            "hooks": Hooks.add,
            "job": Jobs.add,
            "jobs": Jobs.add,
            "files": Files.add,
            "packages": Packages.add,
        }.get(resource, messages.not_implemented)

        add_func(*args, **kwargs)

    @decorators.credentials_check
    def update(self, resource, *args, **kwargs):
        update_func = {
            "dash": Dashes.update,
            "dashes": Dashes.update,
            "form": Forms.update,
            "forms": Forms.update,
            "hook": Hooks.update,
            "hooks": Hooks.update,
            "job": Jobs.update,
            "jobs": Jobs.update,
        }.get(resource, messages.not_implemented)

        update_func(*args, **kwargs)

    @decorators.credentials_check
    def remove(self, resource, *args, **kwargs):
        remove_func = {
            "vars": Vars.remove,
            "dash": Dashes.remove,
            "dashes": Dashes.remove,
            "form": Forms.remove,
            "forms": Forms.remove,
            "hook": Hooks.remove,
            "hooks": Hooks.remove,
            "job": Jobs.remove,
            "jobs": Jobs.remove,
            "files": Files.remove,
            "packages": Packages.remove,
        }.get(resource, messages.not_implemented)

        remove_func(*args, **kwargs)

    @decorators.credentials_check
    def play(self, resource, *args, **kwargs):
        play_func = {
            "dash": Dashes.play,
            "form": Forms.play,
            "hook": Hooks.play,
            "forms": Forms.play,
            "hooks": Hooks.play,
        }.get(resource, messages.not_implemented)

        play_func(*args, **kwargs)

    @decorators.credentials_check
    def deploy(self, **kwargs):
        deploy(**kwargs)

    # Aliases
    def upload(self, *args, **kwargs):
        if not len(args):
            args = ["."]
        self.add("files", *args, **kwargs)

    def ls(self, *args, **kwargs):
        self.list("files", *args, **kwargs)

    def rm(self, *args, **kwargs):
        self.remove("files", *args, **kwargs)

    def install(self, *args, **kwargs):
        self.add("packages", *args, **kwargs)


def _SeparateFlagArgs(args):
    try:
        index = args.index("--help")
        args = args[:index]
        return args, ["--help"]
    except ValueError:
        return args, []


def main():
    fire.core.parser.SeparateFlagArgs = _SeparateFlagArgs
    fire.Fire(CLI)



import os
from abstra_cli.utils import ABSTRA_FOLDER, CREDENTIALS_FILE


def create_abstra_dir():
    if not os.path.exists(ABSTRA_FOLDER):
        os.makedirs(ABSTRA_FOLDER)


def save_credentials(api_token: str):
    create_abstra_dir()
    with open(CREDENTIALS_FILE, "w") as f:
        f.write(api_token.strip())


def get_credentials():
    if os.getenv("ABSTRA_API_TOKEN"):
        return os.getenv("ABSTRA_API_TOKEN")

    if not os.path.exists(CREDENTIALS_FILE):
        return None

    with open(CREDENTIALS_FILE) as f:
        return f.read().strip()



import json
from abstra_cli.resources import Forms, Files, Packages, Vars, Hooks, Jobs, Dashes


ACCEPTED_KEYS = ["files", "workspace", "forms", "hooks", "jobs", "packages", "vars"]


def get_abstra_json_path(parameters: dict) -> str:
    return parameters.get("file") or parameters.get("f") or "abstra.json"


def evaluate_parameters_file(parameters: dict) -> dict:
    file = get_abstra_json_path(parameters)

    data = None
    try:
        with open(file) as f:
            data = json.loads(f.read())
            if not isinstance(data, dict):
                raise Exception("Bad data")
    except:
        raise Exception("Deploy file not found or not correct format")

    for key in data.keys():
        if key not in ACCEPTED_KEYS:
            raise Exception(f"Extra data in deploy file {file} not accepted")

    return data


def deploy(**kwargs):
    deploy_data = evaluate_parameters_file(kwargs)

    for dash_props in Dashes.list_dash_props(get_abstra_json_path(kwargs)):
        Dashes.add(upsert=True, **dash_props)

    forms = deploy_data.pop("forms", None)
    if forms:
        for form in forms:
            Forms.add(upsert=True, **form)

    hooks = deploy_data.pop("hooks", None)
    if hooks:
        for hook in hooks:
            Hooks.add(upsert=True, **hook)

    jobs = deploy_data.pop("jobs", None)
    if jobs:
        for job in jobs:
            Jobs.add(upsert=True, **job)

    files = deploy_data.pop("files", None)
    if files:
        if isinstance(files, dict):
            Files.add(**files)
        elif isinstance(files, list):
            Files.add(*files)

    packages = deploy_data.pop("packages", None)
    if packages:
        if isinstance(packages, dict):
            Packages.add(**packages)
        elif isinstance(packages, list):
            Packages.add(*packages)

    vars = deploy_data.pop("vars", None)
    if vars:
        if isinstance(vars, dict):
            Vars.add(**vars)
        elif isinstance(vars, list):
            Vars.add(*vars)



import webbrowser


from abstra_cli.resources.resources import Resource
import abstra_cli.messages as messages
import abstra_cli.utils as utils
import abstra_cli.apis as apis


NAME_PARAMETERS = ["name", "title"]
PATH_PARAMETERS = ["path"]
CODE_PARAMETERS = ["file", "f", "code", "c"]
BACKGROUND_PARAMETERS = ["background"]
OTHER_PARAMETERS = [
    "start_message",
    "end_message",
    "error_message",
    "timeout_mesage",
    "main_color",
    "start_button_text",
    "restart_button_text",
    "logo_url",
    "log_messages",
    "font_color",
    "welcome_title",
    "brand_name",
]
NON_FLAG_PARAMETERS = (
    NAME_PARAMETERS
    + PATH_PARAMETERS
    + CODE_PARAMETERS
    + BACKGROUND_PARAMETERS
    + OTHER_PARAMETERS
)
FLAG_PARAMETERS = ["auto_start", "allow_restart", "show_sidebar", "enabled"]
FORM_PARAMETERS = FLAG_PARAMETERS + NON_FLAG_PARAMETERS


def check_valid_parameters(parameters: dict) -> None:
    for param in parameters.keys():
        if param not in FORM_PARAMETERS:
            messages.invalid_parameter(param)
            exit()
    for param, value in parameters.items():
        if param in NON_FLAG_PARAMETERS and value in [True, False]:
            messages.invalid_non_flag_parameter_value(param)
            exit()


def evaluate_parameter_name(parameters: dict, use_default=True) -> dict:
    name = parameters.get("name") or parameters.get("n") or parameters.get("title")
    if not name and not use_default:
        return {}
    return {"name": name or "New Form"}


def evaluate_parameter_path(parameters: dict) -> dict:
    path = parameters.get("path")
    if not path:
        return {}
    return {"path": path}


def evaluate_parameters_file_and_code(parameters: dict, use_default=True) -> dict:
    EMPTY_FORM = "from hackerforms import *"
    file = parameters.get("file") or parameters.get("f")
    code = parameters.get("code") or parameters.get("c")

    if file and code:
        messages.code_and_file_not_allowed()
        exit()

    if file:
        with open(file, "r") as f:
            return {"code": f.read()}

    if code:
        return {"code": code}

    return {"code": EMPTY_FORM} if use_default else {}


def evaluate_flag_parameters(parameters: dict) -> dict:
    evaluated_params = {}
    for param, value in parameters.items():
        if param in FLAG_PARAMETERS:
            if value == "true" or value == True:
                evaluated_params[param] = True
                continue
            if value == "false" or value == False:
                evaluated_params[param] = False
                continue

            messages.invalid_flag_parameter_value(param)
            exit()
    return evaluated_params


def evaluate_other_parameters(parameters: dict) -> dict:
    other_parameters = {}
    for param, value in parameters.items():
        if param in OTHER_PARAMETERS:
            other_parameters[param] = value

    return other_parameters


def evaluate_background_parameter_value(parameters: dict) -> dict:
    background = parameters.get("background", None)
    if not background:
        return {}

    if utils.check_color(background):
        return {"theme": background}

    if not utils.path_exists(background):
        messages.file_path_does_not_exists_message(background)
        exit()

    if utils.check_is_image_path(background):
        filename = utils.slugify_filename(background)
        try:
            with open(background, "rb") as f:
                file = f.read()
                url = apis.asset_upload(filename, file)
                return {"theme": url}
        except Exception as e:
            messages.error_upload_background_message(background)
            exit()

    messages.invalid_background_parameter_value()
    exit()


class Forms(Resource):
    @staticmethod
    def list():
        forms = apis.list_workspace_forms()
        messages.print_forms(forms)

    @staticmethod
    def add(*args, **kwargs):
        upsert = kwargs.pop("upsert", False)
        path = kwargs.get("path")
        if upsert and not path:
            messages.upsert_without_identifier("path")
            exit()

        check_valid_parameters(kwargs)

        form_data = {
            **evaluate_parameter_name(kwargs),
            **evaluate_parameter_path(kwargs),
            **evaluate_parameters_file_and_code(kwargs),
            **evaluate_flag_parameters(kwargs),
            **evaluate_other_parameters(kwargs),
            **evaluate_background_parameter_value(kwargs),
        }

        if form_data:
            try:
                if upsert:
                    apis.upsert_workspace_form(form_data)
                    messages.upserted_message("Form", path)
                else:
                    path = apis.add_workspace_form(form_data)["path"]
                    messages.created_message("Form", path)
            except Exception as e:
                print(e)
                messages.create_failed("Form")

    @staticmethod
    def update(*args, **kwargs):
        if not len(args):
            messages.required_argument("path")
            exit()
        path = args[0]

        if not len(kwargs):
            messages.missing_parameters_to_update("form", path)
            exit()

        check_valid_parameters(kwargs)

        form_data = {
            **evaluate_parameter_name(kwargs, use_default=False),
            **evaluate_parameter_path(kwargs),
            **evaluate_parameters_file_and_code(kwargs, use_default=False),
            **evaluate_flag_parameters(kwargs),
            **evaluate_other_parameters(kwargs),
            **evaluate_background_parameter_value(kwargs),
        }

        if form_data:
            try:
                apis.update_workspace_form(path, form_data)
                messages.updated_message("Form", path)
            except Exception as e:
                print(e)
                messages.update_failed("Form", path)

    @staticmethod
    def remove(*args, **kwargs):
        if not len(args):
            messages.required_argument("path")
            exit()

        path = args[0]
        apis.delete_workspace_form(path)
        messages.deleted_message("Form", path)

    @staticmethod
    def play(*args, **kwargs):
        if not len(args):
            messages.required_argument("path")
            exit()

        path = args[0]
        subdomain_name = apis.get_subdomain()
        url = utils.get_prod_form_url(subdomain_name, path)
        messages.form_url(url)
        webbrowser.open(url)
